@@ -3,18 +3,23 @@ import os
 import threading
 import subprocess
 import concurrent.futures
-
-import logbook
 import sys
+import logging
+from packaging import version
 
-from logbook import StreamHandler
-
+from pyshark.packet.packet import Packet
 from pyshark.tshark.tshark import get_process_path, get_tshark_display_filter_flag, \
-    tshark_supports_json, TSharkVersionException
+    tshark_supports_json, TSharkVersionException, get_tshark_version, tshark_supports_duplicate_keys
 from pyshark.tshark.tshark_json import packet_from_json_packet
 from pyshark.tshark.tshark_xml import packet_from_xml_packet, psml_structure_from_xml
 
 
+if sys.version_info < (3, 8):
+    asyncTimeoutError = concurrent.futures.TimeoutError
+else:
+    asyncTimeoutError = asyncio.exceptions.TimeoutError
+
+
 class TSharkCrashException(Exception):
     pass
 
@@ -28,30 +33,27 @@ class RawMustUseJsonException(Exception):
 
 
 class StopCapture(Exception):
-    """
-    Exception that the user can throw anywhere in packet-handling to stop the capture process.
-    """
+    """Exception that the user can throw anywhere in packet-handling to stop the capture process."""
     pass
 
 
-class Capture(object):
-    """
-    Base class for packet captures.
-    """
+class Capture:
+    """Base class for packet captures."""
     DEFAULT_BATCH_SIZE = 2 ** 16
     SUMMARIES_BATCH_SIZE = 64
-    DEFAULT_LOG_LEVEL = logbook.CRITICAL
-    SUPPORTED_ENCRYPTION_STANDARDS = ['wep', 'wpa-pwk', 'wpa-pwd', 'wpa-psk']
+    DEFAULT_LOG_LEVEL = logging.CRITICAL
+    SUPPORTED_ENCRYPTION_STANDARDS = ["wep", "wpa-pwk", "wpa-pwd", "wpa-psk"]
 
     def __init__(self, display_filter=None, only_summaries=False, eventloop=None,
-                 decryption_key=None, encryption_type='wpa-pwd', output_file=None,
+                 decryption_key=None, encryption_type="wpa-pwd", output_file=None,
                  decode_as=None,  disable_protocol=None, tshark_path=None,
-                 override_prefs=None, capture_filter=None, use_json=False, include_raw=False):
+                 override_prefs=None, capture_filter=None, use_json=False, include_raw=False,
+                 custom_parameters=None, debug=False):
 
         self.loaded = False
         self.tshark_path = tshark_path
         self._override_prefs = override_prefs
-        self.debug = False
+        self.debug = debug
         self.use_json = use_json
         self.include_raw = include_raw
         self._packets = []
@@ -63,11 +65,20 @@ class Capture(object):
         self._running_processes = set()
         self._decode_as = decode_as
         self._disable_protocol = disable_protocol
-        self._log = logbook.Logger(self.__class__.__name__, level=self.DEFAULT_LOG_LEVEL)
+        self._json_has_duplicate_keys = True
+        self._log = logging.Logger(
+            self.__class__.__name__, level=self.DEFAULT_LOG_LEVEL)
         self._closed = False
+        self._custom_parameters = custom_parameters
+        self._eof_reached = False
+        self.__tshark_version = None
 
         if include_raw and not use_json:
-            raise RawMustUseJsonException("use_json must be True if include_raw")
+            raise RawMustUseJsonException(
+                "use_json must be True if include_raw")
+
+        if self.debug:
+            self.set_debug()
 
         self.eventloop = eventloop
         if self.eventloop is None:
@@ -76,11 +87,10 @@ class Capture(object):
             self.encryption = (decryption_key, encryption_type.lower())
         else:
             raise UnknownEncyptionStandardException("Only the following standards are supported: %s."
-                                                    % ', '.join(self.SUPPORTED_ENCRYPTION_STANDARDS))
+                                                    % ", ".join(self.SUPPORTED_ENCRYPTION_STANDARDS))
 
     def __getitem__(self, item):
-        """
-        Gets the packet in the given index.
+        """Gets the packet in the given index.
 
         :param item: packet index
         :return: Packet object.
@@ -90,12 +100,12 @@ class Capture(object):
     def __len__(self):
         return len(self._packets)
 
-    def next(self):
+    def next(self) -> Packet:
         return self.next_packet()
 
     # Allows for child classes to call next() from super() without 2to3 "fixing"
     # the call
-    def next_packet(self):
+    def next_packet(self) -> Packet:
         if self._current_packet >= len(self._packets):
             raise StopIteration()
         cur_packet = self._packets[self._current_packet]
@@ -108,14 +118,12 @@ class Capture(object):
         self._current_packet = 0
 
     def reset(self):
-        """
-        Starts iterating packets from the first one.
-        """
+        """Starts iterating packets from the first one."""
         self._current_packet = 0
 
     def load_packets(self, packet_count=0, timeout=None):
-        """
-        Reads the packets from the source (cap, interface, etc.) and adds it to the internal list.
+        """Reads the packets from the source (cap, interface, etc.) and adds it to the internal list.
+
         If 0 as the packet_count is given, reads forever
 
         :param packet_count: The amount of packets to add to the packet list (0 to read forever)
@@ -130,55 +138,86 @@ class Capture(object):
                 raise StopCapture()
 
         try:
-            self.apply_on_packets(keep_packet, timeout=timeout)
+            self.apply_on_packets(
+                keep_packet, timeout=timeout, packet_count=packet_count)
             self.loaded = True
-        except concurrent.futures.TimeoutError:
+        except asyncTimeoutError:
             pass
 
-    def set_debug(self, set_to=True):
-        """
-        Sets the capture to debug mode (or turns it off if specified).
-        """
+    def set_debug(self, set_to=True, log_level=logging.DEBUG):
+        """Sets the capture to debug mode (or turns it off if specified)."""
         if set_to:
-            StreamHandler(sys.stdout).push_application()
-            self._log.level = logbook.DEBUG
+            handler = logging.StreamHandler(sys.stdout)
+            handler.setFormatter(logging.Formatter(
+                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
+            self._log.addHandler(handler)
+            self._log.level = log_level
         self.debug = set_to
 
     def _setup_eventloop(self):
-        """
-        Sets up a new eventloop as the current one according to the OS.
-        """
-        if os.name == 'nt':
+        """Sets up a new eventloop as the current one according to the OS."""
+        if os.name == "nt":
             self.eventloop = asyncio.ProactorEventLoop()
         else:
-            self.eventloop = asyncio.new_event_loop()
-        asyncio.set_event_loop(self.eventloop)
-        if os.name == 'posix' and isinstance(threading.current_thread(), threading._MainThread):
-            asyncio.get_child_watcher().attach_loop(self.eventloop)
+            try:
+                self.eventloop = asyncio.get_event_loop_policy().get_event_loop()
+            except RuntimeError:
+                if threading.current_thread() != threading.main_thread():
+                    # Ran not in main thread, make a new eventloop
+                    self.eventloop = asyncio.new_event_loop()
+                    asyncio.set_event_loop(self.eventloop)
+                else:
+                    raise
+            if os.name == "posix" and isinstance(threading.current_thread(), threading._MainThread):
+                # The default child watchers (ThreadedChildWatcher) attach_loop method is empty!
+                # While using pyshark with ThreadedChildWatcher, asyncio could raise a ChildProcessError
+                # "Unknown child process pid %d, will report returncode 255"
+                # This led to a TSharkCrashException in _cleanup_subprocess.
+                # Using the SafeChildWatcher fixes this issue, but it is slower.
+                # SafeChildWatcher O(n) -> large numbers of processes are slow
+                # ThreadedChildWatcher O(1) -> independent of process number
+                # asyncio.get_child_watcher().attach_loop(self.eventloop)
+                asyncio.set_child_watcher(asyncio.SafeChildWatcher())
+                asyncio.get_child_watcher().attach_loop(self.eventloop)
 
-    @classmethod
-    def _get_json_separator(cls):
-        return ("}%s%s  ," % (os.linesep, os.linesep)).encode()
+    def _get_json_separators(self):
+        """"Returns the separators between packets in a JSON output
 
-    @classmethod
-    def _extract_packet_json_from_data(cls, data, got_first_packet=True):
+        Returns a tuple of (packet_separator, end_of_file_separator, characters_to_disregard).
+        The latter variable being the number of characters to ignore in order to pass the packet (i.e. extra newlines,
+        commas, parenthesis).
+        """
+        if self._get_tshark_version() >= version.parse("3.0.0"):
+            return ("%s  },%s" % (os.linesep, os.linesep)).encode(), ("}%s]" % os.linesep).encode(), (
+                1 + len(os.linesep))
+        else:
+            return ("}%s%s  ," % (os.linesep, os.linesep)).encode(), ("}%s%s]" % (os.linesep, os.linesep)).encode(), 1
+
+    def _extract_packet_json_from_data(self, data, got_first_packet=True):
         tag_start = 0
         if not got_first_packet:
             tag_start = data.find(b"{")
             if tag_start == -1:
                 return None, data
-        closing_tag = cls._get_json_separator()
-        tag_end = data.find(closing_tag)
+        packet_separator, end_separator, end_tag_strip_length = self._get_json_separators()
+        found_separator = None
+
+        tag_end = data.find(packet_separator)
         if tag_end == -1:
-            closing_tag = ("}%s%s]" % (os.linesep, os.linesep)).encode()
-            tag_end = data.find(closing_tag)
-        if tag_end != -1:
-            # Include closing parenthesis but not comma
-            tag_end += len(closing_tag) - 1
-            return data[tag_start:tag_end], data[tag_end + 1:]
+            # Not end of packet, maybe it has end of entire file?
+            tag_end = data.find(end_separator)
+            if tag_end != -1:
+                found_separator = end_separator
+        else:
+            # Found a single packet, just add the separator without extras
+            found_separator = packet_separator
+
+        if found_separator:
+            tag_end += len(found_separator) - end_tag_strip_length
+            return data[tag_start:tag_end].strip().strip(b","), data[tag_end + 1:]
         return None, data
 
-    def _extract_tag_from_data(self, data, tag_name=b'packet'):
+    def _extract_tag_from_data(self, data, tag_name=b"packet"):
         """Gets data containing a (part of) tshark xml.
 
         If the given tag is found in it, returns the tag data and the remaining data.
@@ -187,8 +226,8 @@ class Capture(object):
         :param data: string of a partial tshark xml.
         :return: a tuple of (tag, data). tag will be None if none is found.
         """
-        opening_tag = b'<' + tag_name + b'>'
-        closing_tag = opening_tag.replace(b'<', b'</')
+        opening_tag = b"<" + tag_name + b">"
+        closing_tag = opening_tag.replace(b"<", b"</")
         tag_end = data.find(closing_tag)
         if tag_end != -1:
             tag_end += len(closing_tag)
@@ -197,19 +236,21 @@ class Capture(object):
         return None, data
 
     def _packets_from_tshark_sync(self, packet_count=None, existing_process=None):
-        """
-        Returns a generator of packets.
+        """Returns a generator of packets.
+
         This is the sync version of packets_from_tshark. It wait for the completion of each coroutine and
          reimplements reading packets in a sync way, yielding each packet as it arrives.
 
         :param packet_count: If given, stops after this amount of packets is captured.
         """
         # NOTE: This has code duplication with the async version, think about how to solve this
-        tshark_process = existing_process or self.eventloop.run_until_complete(self._get_tshark_process())
-        psml_structure, data = self.eventloop.run_until_complete(self._get_psml_struct(tshark_process.stdout))
+        tshark_process = existing_process or self.eventloop.run_until_complete(
+            self._get_tshark_process())
+        psml_structure, data = self.eventloop.run_until_complete(
+            self._get_psml_struct(tshark_process.stdout))
         packets_captured = 0
 
-        data = b''
+        data = b""
         try:
             while True:
                 try:
@@ -218,7 +259,8 @@ class Capture(object):
                                                      got_first_packet=packets_captured > 0))
 
                 except EOFError:
-                    self._log.debug('EOF reached (sync)')
+                    self._log.debug("EOF reached (sync)")
+                    self._eof_reached = True
                     break
 
                 if packet:
@@ -227,11 +269,13 @@ class Capture(object):
                 if packet_count and packets_captured >= packet_count:
                     break
         finally:
-            self.eventloop.run_until_complete(self._cleanup_subprocess(tshark_process))
+            if tshark_process in self._running_processes:
+                self.eventloop.run_until_complete(
+                    self._cleanup_subprocess(tshark_process))
 
     def apply_on_packets(self, callback, timeout=None, packet_count=None):
-        """
-        Runs through all packets and calls the given callback (a function) with each one as it is read.
+        """Runs through all packets and calls the given callback (a function) with each one as it is read.
+
         If the capture is infinite (i.e. a live capture), it will run forever, otherwise it will complete after all
         packets have been read.
 
@@ -261,13 +305,12 @@ class Capture(object):
             pass
         finally:
             if close_tshark:
-                await self._close_async()
-                #yield From(self._cleanup_subprocess(tshark_process))
+                await self.close_async()
 
     async def _go_through_packets_from_fd(self, fd, packet_callback, packet_count=None):
         """A coroutine which goes through a stream and calls a given callback for each XML packet seen in it."""
         packets_captured = 0
-        self._log.debug('Starting to go through packets')
+        self._log.debug("Starting to go through packets")
 
         psml_struct, data = await self._get_psml_struct(fd)
 
@@ -276,7 +319,8 @@ class Capture(object):
                 packet, data = await self._get_packet_from_stream(fd, data, got_first_packet=packets_captured > 0,
                                                                   psml_structure=psml_struct)
             except EOFError:
-                self._log.debug('EOF reached')
+                self._log.debug("EOF reached")
+                self._eof_reached = True
                 break
 
             if packet:
@@ -284,7 +328,7 @@ class Capture(object):
                 try:
                     packet_callback(packet)
                 except StopCapture:
-                    self._log.debug('User-initiated capture stop in callback')
+                    self._log.debug("User-initiated capture stop in callback")
                     break
 
             if packet_count and packets_captured >= packet_count:
@@ -296,7 +340,7 @@ class Capture(object):
 
         A coroutine.
         """
-        data = b''
+        data = b""
         psml_struct = None
 
         if self._only_summaries:
@@ -304,7 +348,8 @@ class Capture(object):
             while not psml_struct:
                 new_data = await fd.read(self.SUMMARIES_BATCH_SIZE)
                 data += new_data
-                psml_struct, data = self._extract_tag_from_data(data, b'structure')
+                psml_struct, data = self._extract_tag_from_data(
+                    data, b"structure")
                 if psml_struct:
                     psml_struct = psml_structure_from_xml(psml_struct)
                 elif not new_data:
@@ -329,9 +374,11 @@ class Capture(object):
 
         if packet:
             if self.use_json:
-                packet = packet_from_json_packet(packet)
+                packet = packet_from_json_packet(
+                    packet, deduplicate_fields=self._json_has_duplicate_keys)
             else:
-                packet = packet_from_xml_packet(packet, psml_structure=psml_structure)
+                packet = packet_from_xml_packet(
+                    packet, psml_structure=psml_structure)
             return packet, existing_data
 
         new_data = await stream.read(self.DEFAULT_BATCH_SIZE)
@@ -339,6 +386,7 @@ class Capture(object):
 
         if not new_data:
             # Reached EOF
+            self._eof_reached = True
             raise EOFError()
         return None, existing_data
 
@@ -347,23 +395,32 @@ class Capture(object):
 
     def _stderr_output(self):
         # Ignore stderr output unless in debug mode (sent to console)
-        return None if self.debug else open(os.devnull, "w")
+        return None if self.debug else subprocess.DEVNULL
+
+    def _get_tshark_version(self):
+        if self.__tshark_version is None:
+            self.__tshark_version = get_tshark_version(self.tshark_path)
+        return self.__tshark_version
 
     async def _get_tshark_process(self, packet_count=None, stdin=None):
-        """
-        Returns a new tshark process with previously-set parameters.
-        """
+        """Returns a new tshark process with previously-set parameters."""
+        output_parameters = []
         if self.use_json:
-            output_type = 'json'
-            if not tshark_supports_json(self.tshark_path):
-                raise TSharkVersionException("JSON only supported on Wireshark >= 2.2.0")
+            output_type = "json"
+            if not tshark_supports_json(self._get_tshark_version()):
+                raise TSharkVersionException(
+                    "JSON only supported on Wireshark >= 2.2.0")
+            if tshark_supports_duplicate_keys(self._get_tshark_version()):
+                output_parameters.append("--no-duplicate-keys")
+                self._json_has_duplicate_keys = False
         else:
-            output_type = 'psml' if self._only_summaries else 'pdml'
-        parameters = [self._get_tshark_path(), '-l', '-n', '-T', output_type] + \
-                      self.get_parameters(packet_count=packet_count)
-
-        self._log.debug('Creating TShark subprocess with parameters: ' + ' '.join(parameters))
+            output_type = "psml" if self._only_summaries else "pdml"
+        parameters = [self._get_tshark_path(), "-l", "-n", "-T", output_type] + \
+            self.get_parameters(packet_count=packet_count) + output_parameters
 
+        self._log.debug(
+            "Creating TShark subprocess with parameters: " + " ".join(parameters))
+        self._log.debug("Executable: %s" % parameters[0])
         tshark_process = await asyncio.create_subprocess_exec(*parameters,
                                                               stdout=subprocess.PIPE,
                                                               stderr=self._stderr_output(),
@@ -372,38 +429,39 @@ class Capture(object):
         return tshark_process
 
     def _created_new_process(self, parameters, process, process_name="TShark"):
-        self._log.debug('%s subprocess created', process_name)
+        self._log.debug(
+            process_name + f" subprocess (pid {process.pid}) created")
         if process.returncode is not None and process.returncode != 0:
             raise TSharkCrashException(
-                '%s seems to have crashed. Try updating it. (command ran: "%s")' % (
-                    process_name, ' '.join(parameters)))
+                "%s seems to have crashed. Try updating it. (command ran: '%s')" % (
+                    process_name, " ".join(parameters)))
         self._running_processes.add(process)
 
     async def _cleanup_subprocess(self, process):
-        """
-        Kill the given process and properly closes any pipes connected to it.
-        """
+        """Kill the given process and properly closes any pipes connected to it."""
+        self._log.debug(f"Cleanup Subprocess (pid {process.pid})")
         if process.returncode is None:
             try:
                 process.kill()
                 return await asyncio.wait_for(process.wait(), 1)
-            except concurrent.futures.TimeoutError:
-                self._log.debug('Waiting for process to close failed, may have zombie process.')
+            except asyncTimeoutError:
+                self._log.debug(
+                    "Waiting for process to close failed, may have zombie process.")
             except ProcessLookupError:
                 pass
             except OSError:
-                if os.name != 'nt':
+                if os.name != "nt":
                     raise
         elif process.returncode > 0:
-            raise TSharkCrashException('TShark seems to have crashed (retcode: %d). '
-                                       'Try rerunning in debug mode [ capture_obj.set_debug() ] or try updating tshark.'
-                                       % process.returncode)
+            if process.returncode != 1 or self._eof_reached:
+                raise TSharkCrashException(f"TShark (pid {process.pid}) seems to have crashed (retcode: {process.returncode}). "
+                                           "Try rerunning in debug mode [ capture_obj.set_debug() ] or try updating tshark.")
 
     def close(self):
-        self.eventloop.run_until_complete(self._close_async())
+        self.eventloop.create_task(self.close_async())
 
-    async def _close_async(self):
-        for process in self._running_processes:
+    async def close_async(self):
+        for process in self._running_processes.copy():
             await self._cleanup_subprocess(process)
         self._running_processes.clear()
 
@@ -411,38 +469,56 @@ class Capture(object):
         if self._running_processes:
             self.close()
 
+    def __enter__(self): return self
+    async def __aenter__(self): return self
+    def __exit__(self, exc_type, exc_val, exc_tb): self.close()
+
+    async def __aexit__(self, exc_type, exc_val,
+                        exc_tb): await self.close_async()
+
     def get_parameters(self, packet_count=None):
-        """
-        Returns the special tshark parameters to be used according to the configuration of this class.
-        """
+        """Returns the special tshark parameters to be used according to the configuration of this class."""
         params = []
         if self._capture_filter:
-            params += ['-f', self._capture_filter]
+            params += ["-f", self._capture_filter]
         if self._display_filter:
-            params += [get_tshark_display_filter_flag(self.tshark_path), self._display_filter]
+            params += [get_tshark_display_filter_flag(self._get_tshark_version(),),
+                       self._display_filter]
         # Raw is only enabled when JSON is also enabled.
         if self.include_raw:
             params += ["-x"]
         if packet_count:
-            params += ['-c', str(packet_count)]
+            params += ["-c", str(packet_count)]
+
+        if self._custom_parameters:
+            if isinstance(self._custom_parameters, list):
+                params += self._custom_parameters
+            elif isinstance(self._custom_parameters, dict):
+                for key, val in self._custom_parameters.items():
+                    params += [key, val]
+            else:
+                raise TypeError("Custom parameters type not supported.")
+
         if all(self.encryption):
-            params += ['-o', 'wlan.enable_decryption:TRUE', '-o', 'uat:80211_keys:"' + self.encryption[1] + '","' +
+            params += ["-o", "wlan.enable_decryption:TRUE", "-o", 'uat:80211_keys:"' + self.encryption[1] + '","' +
                                                                   self.encryption[0] + '"']
         if self._override_prefs:
             for preference_name, preference_value in self._override_prefs.items():
-                if all(self.encryption) and preference_name in ('wlan.enable_decryption', 'uat:80211_keys'):
+                if all(self.encryption) and preference_name in ("wlan.enable_decryption", "uat:80211_keys"):
                     continue  # skip if override preferences also given via --encryption options
-                params += ['-o', '{0}:{1}'.format(preference_name, preference_value)]
+                params += ["-o",
+                           "{0}:{1}".format(preference_name, preference_value)]
 
         if self._output_file:
-            params += ['-w', self._output_file]
+            params += ["-w", self._output_file]
 
         if self._decode_as:
             for criterion, decode_as_proto in self._decode_as.items():
-                params += ['-d', ','.join([criterion.strip(), decode_as_proto.strip()])]
+                params += ["-d",
+                           ",".join([criterion.strip(), decode_as_proto.strip()])]
 
         if self._disable_protocol:
-            params += ['--disable-protocol', self._disable_protocol.strip()]
+            params += ["--disable-protocol", self._disable_protocol.strip()]
 
         return params
 
@@ -453,4 +529,4 @@ class Capture(object):
             return self._packets_from_tshark_sync()
 
     def __repr__(self):
-        return '<%s (%d packets)>' % (self.__class__.__name__, len(self._packets))
+        return "<%s (%d packets)>" % (self.__class__.__name__, len(self._packets))
