@@ -1,8 +1,13 @@
-import pyshark
-import py
 import os
 
+from configparser import ConfigParser
+
+import pyshark
+
 CONFIG_PATH = os.path.join(os.path.dirname(pyshark.__file__), 'config.ini')
 
+
 def get_config():
-    return py.iniconfig.IniConfig(CONFIG_PATH)
+    config = ConfigParser()
+    config.read(CONFIG_PATH)
+    return config
